"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
module.exports = wrapAnsi;

var _stringWidth = _interopRequireDefault(require("string-width"));

var _stripAnsi = _interopRequireDefault(require("strip-ansi"));

var _ansiStyles = _interopRequireDefault(require("ansi-styles"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }

function _iterableToArrayLimit(arr, i) { var _i = arr == null ? null : typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"]; if (_i == null) return; var _arr = []; var _n = true; var _d = false; var _s, _e; try { for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _createForOfIteratorHelper(o, allowArrayLike) { var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"]; if (!it) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e2) { throw _e2; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = it.call(o); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e3) { didErr = true; err = _e3; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _iterableToArray(iter) { if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) return _arrayLikeToArray(arr); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var ESCAPES = new Set(["\x1B", "\x9B"]);
var END_CODE = 39;
var ANSI_ESCAPE_BELL = "\x07";
var ANSI_CSI = '[';
var ANSI_OSC = ']';
var ANSI_SGR_TERMINATOR = 'm';
var ANSI_ESCAPE_LINK = "".concat(ANSI_OSC, "8;;");

var wrapAnsiCode = function wrapAnsiCode(code) {
  return "".concat(ESCAPES.values().next().value).concat(ANSI_CSI).concat(code).concat(ANSI_SGR_TERMINATOR);
};

var wrapAnsiHyperlink = function wrapAnsiHyperlink(uri) {
  return "".concat(ESCAPES.values().next().value).concat(ANSI_ESCAPE_LINK).concat(uri).concat(ANSI_ESCAPE_BELL);
}; // Calculate the length of words split on ' ', ignoring
// the extra characters added by ansi escape codes


var wordLengths = function wordLengths(string) {
  return string.split(' ').map(function (character) {
    return (0, _stringWidth["default"])(character);
  });
}; // Wrap a long word across multiple rows
// Ansi escape codes do not count towards length


var wrapWord = function wrapWord(rows, word, columns) {
  var characters = _toConsumableArray(word);

  var isInsideEscape = false;
  var isInsideLinkEscape = false;
  var visible = (0, _stringWidth["default"])((0, _stripAnsi["default"])(rows[rows.length - 1]));

  var _iterator = _createForOfIteratorHelper(characters.entries()),
      _step;

  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var _step$value = _slicedToArray(_step.value, 2),
          index = _step$value[0],
          character = _step$value[1];

      var characterLength = (0, _stringWidth["default"])(character);

      if (visible + characterLength <= columns) {
        rows[rows.length - 1] += character;
      } else {
        rows.push(character);
        visible = 0;
      }

      if (ESCAPES.has(character)) {
        isInsideEscape = true;
        isInsideLinkEscape = characters.slice(index + 1).join('').startsWith(ANSI_ESCAPE_LINK);
      }

      if (isInsideEscape) {
        if (isInsideLinkEscape) {
          if (character === ANSI_ESCAPE_BELL) {
            isInsideEscape = false;
            isInsideLinkEscape = false;
          }
        } else if (character === ANSI_SGR_TERMINATOR) {
          isInsideEscape = false;
        }

        continue;
      }

      visible += characterLength;

      if (visible === columns && index < characters.length - 1) {
        rows.push('');
        visible = 0;
      }
    } // It's possible that the last row we copy over is only
    // ansi escape characters, handle this edge-case

  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }

  if (!visible && rows[rows.length - 1].length > 0 && rows.length > 1) {
    rows[rows.length - 2] += rows.pop();
  }
}; // Trims spaces from a string ignoring invisible sequences


var stringVisibleTrimSpacesRight = function stringVisibleTrimSpacesRight(string) {
  var words = string.split(' ');
  var last = words.length;

  while (last > 0) {
    if ((0, _stringWidth["default"])(words[last - 1]) > 0) {
      break;
    }

    last--;
  }

  if (last === words.length) {
    return string;
  }

  return words.slice(0, last).join(' ') + words.slice(last).join('');
}; // The wrap-ansi module can be invoked in either 'hard' or 'soft' wrap mode
//
// 'hard' will never allow a string to take up more than columns characters
//
// 'soft' allows long words to expand past the column length


var exec = function exec(string, columns) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (options.trim !== false && string.trim() === '') {
    return '';
  }

  var returnValue = '';
  var escapeCode;
  var escapeUrl;
  var lengths = wordLengths(string);
  var rows = [''];

  var _iterator2 = _createForOfIteratorHelper(string.split(' ').entries()),
      _step2;

  try {
    for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
      var _step2$value = _slicedToArray(_step2.value, 2),
          index = _step2$value[0],
          word = _step2$value[1];

      if (options.trim !== false) {
        rows[rows.length - 1] = rows[rows.length - 1].trimStart();
      }

      var rowLength = (0, _stringWidth["default"])(rows[rows.length - 1]);

      if (index !== 0) {
        if (rowLength >= columns && (options.wordWrap === false || options.trim === false)) {
          // If we start with a new word but the current row length equals the length of the columns, add a new row
          rows.push('');
          rowLength = 0;
        }

        if (rowLength > 0 || options.trim === false) {
          rows[rows.length - 1] += ' ';
          rowLength++;
        }
      } // In 'hard' wrap mode, the length of a line is never allowed to extend past 'columns'


      if (options.hard && lengths[index] > columns) {
        var remainingColumns = columns - rowLength;
        var breaksStartingThisLine = 1 + Math.floor((lengths[index] - remainingColumns - 1) / columns);
        var breaksStartingNextLine = Math.floor((lengths[index] - 1) / columns);

        if (breaksStartingNextLine < breaksStartingThisLine) {
          rows.push('');
        }

        wrapWord(rows, word, columns);
        continue;
      }

      if (rowLength + lengths[index] > columns && rowLength > 0 && lengths[index] > 0) {
        if (options.wordWrap === false && rowLength < columns) {
          wrapWord(rows, word, columns);
          continue;
        }

        rows.push('');
      }

      if (rowLength + lengths[index] > columns && options.wordWrap === false) {
        wrapWord(rows, word, columns);
        continue;
      }

      rows[rows.length - 1] += word;
    }
  } catch (err) {
    _iterator2.e(err);
  } finally {
    _iterator2.f();
  }

  if (options.trim !== false) {
    rows = rows.map(function (row) {
      return stringVisibleTrimSpacesRight(row);
    });
  }

  var pre = _toConsumableArray(rows.join('\n'));

  var _iterator3 = _createForOfIteratorHelper(pre.entries()),
      _step3;

  try {
    for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
      var _step3$value = _slicedToArray(_step3.value, 2),
          _index = _step3$value[0],
          character = _step3$value[1];

      returnValue += character;

      if (ESCAPES.has(character)) {
        var _ref = new RegExp("(?:\\".concat(ANSI_CSI, "(?<code>\\d+)m|\\").concat(ANSI_ESCAPE_LINK, "(?<uri>.*)").concat(ANSI_ESCAPE_BELL, ")")).exec(pre.slice(_index).join('')) || {
          groups: {}
        },
            groups = _ref.groups;

        if (groups.code !== undefined) {
          var _code = Number.parseFloat(groups.code);

          escapeCode = _code === END_CODE ? undefined : _code;
        } else if (groups.uri !== undefined) {
          escapeUrl = groups.uri.length === 0 ? undefined : groups.uri;
        }
      }

      var code = _ansiStyles["default"].codes.get(Number(escapeCode));

      if (pre[_index + 1] === '\n') {
        if (escapeUrl) {
          returnValue += wrapAnsiHyperlink('');
        }

        if (escapeCode && code) {
          returnValue += wrapAnsiCode(code);
        }
      } else if (character === '\n') {
        if (escapeCode && code) {
          returnValue += wrapAnsiCode(escapeCode);
        }

        if (escapeUrl) {
          returnValue += wrapAnsiHyperlink(escapeUrl);
        }
      }
    }
  } catch (err) {
    _iterator3.e(err);
  } finally {
    _iterator3.f();
  }

  return returnValue;
}; // For each newline, invoke the method separately


function wrapAnsi(string, columns, options) {
  return String(string).normalize().replace(/\r\n/g, '\n').split('\n').map(function (line) {
    return exec(line, columns, options);
  }).join('\n');
}
